@@ -2,10 +2,7 @@ from django.core.management.base import BaseCommand, CommandError
 from computedfields.models import ComputedFieldsModelType
 from django.conf import settings
 from computedfields.graph import ComputedModelsGraph
-try:
-    from django.utils.six.moves import cPickle as pickle
-except ImportError:
-    import pickle
+import pickle
 
 
 class Command(BaseCommand):
@@ -17,8 +14,11 @@ class Command(BaseCommand):
 
         with open(settings.COMPUTEDFIELDS_MAP, 'wb') as f:
             graph = ComputedModelsGraph(ComputedFieldsModelType._computed_models)
-            graph.remove_redundant()
+            if not getattr(settings, 'COMPUTEDFIELDS_ALLOW_RECURSION', False):
+                graph.remove_redundant()
+                graph.get_uniongraph().get_edgepaths()  # uniongraph cyclefree?
             pickle.dump({
                 'lookup_map': graph.generate_lookup_map(),
-                'fk_map': graph._fk_map
+                'fk_map': graph._fk_map,
+                'local_mro': graph.generate_local_mro_map()  # also tests for cycles on modelgraphs
             }, f, pickle.HIGHEST_PROTOCOL)
